#!/usr/bin/python
import sys
import os
import rlcompleter
import readline  
import subprocess
from datetime import date
from datetime import timedelta
from docopt import docopt
 
readline.parse_and_bind("tab: complete")

######################################################
# the functions which invoke subprocess

def get_find_results( directory, option = "" ):
    cmd =""" find %s %s 2>/dev/null"""%( directory, option )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split()
    return results


def do_ls( source, option = ""  ):
    # note: if source does not exist, this function returns an empty list.
    cmd =""" ls %s %s 2>/dev/null"""%( source, option )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split()

    return results


def do_cp( sources, dir ):
    ''' input: sources is a list contain the files to be copied'''

    src_str = ""
    for source in sources: 
        src_str += source

    cmd = "cp -ar %s %s" %( src_str, dir )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE ) 

 
def do_rsync_dry_run( source, dirBackup ):
    '''call rsync --dry-run '''

    cmd =""" rsync --verbose --compress \
                   --archive --recursive \
                   --itemize-changes --delete \
                   --dry-run  --update %s %s 
    """%( source, dirBackup )
    
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split("\n") 
    results = results[1:-3]      
    return results


 
def do_rsync( source, dirBackup, port=22  ):
     
    cmd = ""
    if( port == 22 ):
        # for local and ssh with normal port =22
        cmd =""" rsync --verbose --compress \
                       --archive --recursive \
                       --itemize-changes --delete \
                       --update %s %s 
        """%( source, dirBackup )
    else:
        # for ssh with port other than 22
        cmd =""" rsync --verbose --compress \
                 --archive --recursive \
                 --itemize-changes --delete \
                 -e "ssh -p %d -o 'ControlPath=$HOME/.ssh/ctl/%%r@%%h:%%p'" \
                 --update  %s %s 
        """%( port, source, dirBackup )    
         
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
     

    
def do_cat( fName ):
    cmd =""" cat  %s """  %( fName )
    subprocess.check_call( cmd, shell=True  )


def getAbsolutePath_from_relative( relative_path ):
    # convert relative path to absolute path
    # it use 'dirname' and 'basename'
    cmd=""" echo "$(cd "$(dirname %s)"; pwd)/$(basename %s)"  """ \
         %( relative_path, relative_path)
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
    abs_path=result.strip()
    return abs_path 


def getDateTime_info( ):
    cmd = 'date "+%Y%m%d%H%M"'
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (dateTimeInfo,err) = p.communicate() 
    dateTimeInfo = dateTimeInfo.strip()
    return dateTimeInfo


def getCurrentPath( ):
    cmd="pwd"
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
    currentPATH=result.strip()
    return currentPATH 


def makePatch( old, new , patch_fileName ):
    # note -u option give us time stamp
    cmd = "diff -u %s %s > %s" %( old, new, patch_fileName )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate()
    

def isDifferent( old, new ):
    cmd = "diff -u %s %s" %( old, new )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
     
    if result != "":
        return True
    else:
        return False




#### the following functions do not have return

def applyPatch_Reverse( fName, patch_fName ):
    
    # option "--reject-file -" or "-r -" is to discard .rej files.
    cmd = "patch -r - -R --ignore-whitespace %s < %s " %(fName, patch_fName )
    
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate()

    # print 'debug result = ', result 
    # print 'debug err = ', err 

   
    

def make_a_fold( folderName ):
    cmd = "mkdir -p %s" %( folderName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )

def remove_a_folder( folderName ):
    cmd = "rm -r %s" %( folderName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
    pass


def do_touch( dateTime, fName ):
    # change modification time of a file
    cmd = "touch  -m -t %s %s" %( dateTime, fName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
    pass

##### handy functions (no subprocess )

def get_fileName( _fPath ):
    # extract file name from a path
    
    if( _fPath.find("/") == -1 ):
        return _fPath

    N = len(_fPath)
    idx = 0 
    for i in range( len(_fPath) ):
        if( _fPath[ -N + i ] == '/' ):
            idx = i 
              
    return _fPath[ idx+1: ] 


def get_dirName( _fileName ):
     
    N = len(_fileName)
    idx = 0 
    for i in range( len(_fileName) ):
        if( _fileName[ -N + i ] == '/' ):
            idx = -N + i 
              
    return _fileName[ :idx ] 

#=======================================================================




class RSYNC( ):
    """ to use rsync to do back up and manage version history """
    
    def __init__( self ):
        
        # the backupDir is the topmost dir for backup
        # sources can be either dir or files.
        # they can be assigned by set_main_backup_folder 
        # and add_source methods
        self.backupDir = ""
        self.backupDir_remote=""
        self.port=22
        self.sources = []
        self.do_local = True
        self.do_remote = False

        self.upperDays = sys.maxint 
        # keep version upto XX days.
        # it can be changed by set_upper_days()

        # for processing
        self.modified_files = []
        self.deleted_files = []
        self.history_versions = []
        self.dry_run_results = []

        # fixed dirname
        self.tempDir=".temp"        
        self.historyDir=".history"  
        pass

    def debug_run( self ):
        
        self.get_run_dry_result( debug = True )

        self.get_deleted_files( debug = True )        
        if( len( self.deleted_files ) > 0 ):
            self.archieve_deleted_files_entrieCopy( debug = True ) 

        self.get_modified_files( debug = True )
        if( len( self.modified_files )>0 ):
            self.make_patches_for_modified_files( debug = True )  



    def process( self ):
        ''' the main backup process is done here'''
        
        # self.debug_run() # testing purpose.

        self.get_run_dry_result(  ) 
        if len(self.dry_run_results) == 0 :
            print "up to date, no need to backup"
            return 

        self.get_deleted_files()  
        if( len( self.deleted_files ) > 0 ):
            self.archieve_deleted_files_entrieCopy()  

        self.get_modified_files()  
        if( len( self.modified_files )>0 ):
            self.make_patches_for_modified_files()  
        
        self.manage_history_version()

        if( self.do_local ):
            self.update_backupDir( "local" )
            print "backup done" 
        
        if( self.do_remote ):
            self.update_backupDir( "remote" )
            print "backup remote done" 
            pass
        

        
        pass

    ###################################################
    # methods related to user setting

    def __parse( self, line, key ):
        line = line.strip() 
        STATUSLEN = len(key) + 1
        item = line[ STATUSLEN+1: ] 
        item = item.strip()
        return item

    def read_in_backup_plan( self, backup_plan ):
        ''' read in a backup instruction and then process it '''
        
        with open(backup_plan, "r" ) as f:
            lines = f.readlines()

        for line in lines:
            if( line[0] == "#" or len(line) == 0 ) : continue

            if line.find( "local_backup_folder" ) != -1:
                self.backupDir = self.__parse( line, "local_backup_folder" )
                
            elif line.find( "remote_backup_folder" ) != -1:
                self.backupDir_remote = self.__parse( line, "remote_backup_folder" )
                 
            elif line.find( "ssh_port" ) != -1:
                self.port =  int( self.__parse(line, "ssh_port") ) 
                 
            elif line.find( "add_source" ) != -1 :
                self.add_source( self.__parse(line, "add_source") )
            
            elif line.find( "keep_version" ) != -1 :     
                duration = int( self.__parse(line, "keep_version") ) 
                self.upperDays = duration 
            
            elif line.find( "do_local_backup" ) != -1 :     
                flag = self.__parse(line, "do_local_backup") 
                if flag.lower() ==  "true":
                    self.do_local = True
                else:
                    self.do_local = False

            elif line.find( "do_remote_backup" ) != -1 :     
                flag = self.__parse(line, "do_remote_backup") 
                if flag.lower() ==  "true":
                    self.do_remote = True
                else:
                    self.do_remote = False
                  
        pass

 
    def add_source( self, fileName_path ):
        ''' add to self.sources'''
        self.sources.append( fileName_path )

    ###################################################
    # 

    def validate_fName_dateTime( self ):
        fName = raw_input("input filename: ")
        versions = self.get_file_versions( fName )
        if( len(versions) == 0 ): return

        print "available time stamp: "
        for v in versions: 
            print self.convert_to_readable( v )
        
        while( 1 ):
            dateTime = raw_input("input time stamp: " )
            dateTime = self.convert_to_machine( dateTime)
            if dateTime not in versions:
                print "%s is not available for %s" \
                %( self.convert_to_readable( dateTime ), fName )
            else:
                break
        return fName, dateTime 


    def validate_fName_dateTime2( self ):
        ''' for deleted files '''
        historyDir = self.backupDir + "/" + self.historyDir + "/"
        results = get_find_results( historyDir, "-type f" )

        fName = raw_input("input filename: ")
        if( fName[:2] == './' ): fName.replace("./", "" )
         
        
        notFound = True
        available_dateTime = []
        for result in results:
            if result.find( fName) != -1:
                notFound = False
                dateTime = result.replace( historyDir ,"" )
                dateTime = dateTime.split("/")[0]
                available_dateTime.append( dateTime )
                print "%s %s" \
                %( self.convert_to_readable( dateTime), fName )
                  
        if notFound : 
            print "%s not in the deleted-file list" %(fName)
            return 0
        

       
        if( len(available_dateTime) == 1 ):
            fName = historyDir  + available_dateTime[0] + "/deleted/" + fName
            return fName 
        else:
            print "select time stemp: "

            for v in available_dateTime: 
                print self.convert_to_readable( v )
        
            while( 1 ):
                dateTime = raw_input("input time stamp: " )
                dateTime = self.convert_to_machine( dateTime )
                if dateTime not in available_dateTime:
                    print "%s is not available for %s" \
                    %( self.convert_to_readable( dateTime ), fName )
                else:
                    fName = historyDir  + available_dateTime[0] + "/deleted/" + fName
                    break
            return fName 


    def actions( self, opt ):
        if( opt.lower() == "xx" ):
            sys.exit(1) 

        elif( opt == "1" ):
            versions = self.get_history_versions()
            for v in versions: 
                print self.convert_to_readable( v )
        
        elif( opt == "2" ):
            fName = raw_input("input filename: ")
            versions = self.get_file_versions( fName )
            for v in versions: 
                print self.convert_to_readable( v )
        
        elif( opt == "3" ):
            (fName, dateTime) = self.validate_fName_dateTime()
            self.view_previous_version( fName, dateTime )

        elif( opt == "4" ):
            (fName, dateTime) = self.validate_fName_dateTime()
            self.back_to_previous_version( fName, dateTime )

        elif( opt == "5" ):
            self.show_all_deleted_files( ) 

        elif( opt == "6" ): 
            fName = self.validate_fName_dateTime2()
            if( fName != 0   ):
                print "preview ", fName, ": "
                do_cat( fName )

        elif( opt == "7" ): 
            fName = self.validate_fName_dateTime2()
            if( fName != 0   ):

                fName_source = fName.split( "/deleted/")[-1]
                source_dir = get_dirName( fName_source )
                do_cp( fName, source_dir )
                print "recover: %s" %fName_source

         


        raw_input( "\npress any key to continue")
        os.system('clear');        

    def show_menu( self ):
        '''
        print the menu and then return the option.
        '''
        os.system('clear');
         
        strOut =""
        strOut+="""
    -----------------------------------------------------------
    (1) Show all time stamps
    (2) Show all time stamps for a given file
    (3) View a file by a time stamp
    (4) Roll back a file back to a time stamp
    (5) Show all deleted files
    (6) View a deleted file 
    (7) Recover a deleted file.
    -----------------------------------------------------------
    (XX) Exit  

    Your choice:  """ 
        
        while (1):    
            opt = raw_input( strOut )
            self.actions( opt )      
           
        
         

    ###################################################
    # methods related to retrieve info
    
    def get_modified_files( self, debug = False ):
        ''' this function return a list containing the filenames for 
            modified files.
        '''

        ## warning message
        if self.backupDir == "":
            print "no backup folder is set"
            print "use set_main_backup_folder() "
            return 

        ## warning message    
        if len( self.sources ) == 0 :
            print "no source is set"
            print "use add_source() "
            return 

        STATUSLEN = 12
        for result in self.dry_run_results:
            if( result[:2] == ">f" and (result.find("+") == -1 ) ):
                filename = result[ STATUSLEN: ] 
                if( filename not in self.modified_files ):
                    self.modified_files.append( filename )

        #---------------------------
        if( debug ):
            print "inside get_modified_files()"
            print "the following are modified files:"
            for item in self.modified_files:
                print "modified file:", item
            print "end\n"
        #---------------------------
             
        return self.modified_files
          



    def get_deleted_files( self , debug=False):
        ''' this function return a list containing the filenames for 
            deleted files.
        '''
        ## warning message
        if self.backupDir == "":
            print "no backup folder is set"
            print "use set_main_backup_folder() "
            return 

        ## warning message    
        if len( self.sources ) == 0 :
            print "no source is set"
            print "use add_source() "
            return 

        STATUSLEN = 12    
        for result in self.dry_run_results:
            if(  result.find("deleting") != -1  ):
                filename = result[ STATUSLEN: ] 
                self.deleted_files.append( filename )


        #---------------------------
        if( debug ):
            print "inside get_deleted_files()"
            print "TEST backupDir = ", self.backupDir
            print "TEST sources ", self.sources
            print "The following are deleted files:"
            for item in self.deleted_files:
                print "deleted_file:", item
            print "end\n"
        #---------------------------


        return  self.deleted_files
       
 
     



    ###################################################
    # methods related to do archieve and patch  


    def update_backupDir( self, flag ):
        
        if( flag == "local" ):
            src = ""
            for source in self.sources:
                src += source + " "
            do_rsync( src, self.backupDir, port=22 ) 

        if( flag == "remote" ):    
            do_rsync( self.backupDir + "/" , self.backupDir_remote, port=self.port )
        
   



    def archieve_deleted_files_entrieCopy( self, debug = False ):
        ''' to backup the deleted files 
            we will make a local copy first, then push to remote.
        '''
        
        deletedFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/deleted/"
        
        if( debug):
            print "inside archieve_deleted_files_entrieCopy()"
            print "deletedFolder:", deletedFolder
            print "=> copy deleted files to deletedFolder"

        for source in self.deleted_files:
            dirName = get_dirName( deletedFolder + source )
            make_a_fold( dirName )
            src = self.backupDir+"/"+ source    
            # do_rsync( src, dirName )    
            do_cp( src, dirName )    
            if( debug):
                print "we create a folder:", dirName
        
        if( debug):
            print "end\n"
          
        
        pass


    def archieve_modified_files_entrieCopy( self ):
        ''' currently not in used...''' 
        archiveFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/changed/"
        
        make_a_fold( archiveFolder )                
        src = ""
        for source in self.modified_files:
            src += self.backupDir+"/"+ source + " "
        
        do_rsync( src, archiveFolder )    
        pass


    def make_patches_for_modified_files( self, debug = False ):
        ''' we will make patch files 
        '''
        # note: rsync uses modified date to compare
        # so even the content doesn't change, it is possible
        # to be considered changed file.


        patchFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/patch/"
        
        if( debug ):
            print "inside make_patches_for_modified_files()"
            print "patchFolder:", patchFolder  

        for source in self.modified_files:
            old_file = self.backupDir+ "/" + source
            new_file = source 
            path_file = patchFolder + source + ".patch"

            if( isDifferent( old_file, new_file ) ):
                dirName = get_dirName( patchFolder + source )
                make_a_fold( dirName )
                makePatch( old_file, new_file , path_file )

                if( debug ):
                    print "pathfile:" , path_file, \
                           "old=", old_file, \
                           "new=", new_file
        pass


    ###################################################
    # methods related to modified files



    def get_file_versions( self, fName ):
        ''' given a file name, it returns modifying history ''' 
        
        if( fName[:2] == "./" ):
            fName = fName[2:]

        # check input 
        if len( do_ls(fName) )== 0:
            print "cannot find %s" %fName
            return 

        
         
        historyDir = self.backupDir + "/" +  self.historyDir
        results = get_find_results( historyDir ) 
         
        idx1 = len( historyDir) + 1
        idx2 = idx1 + 12 
        versions = []
        for result in results:
            if( result.find( fName) != -1 and result.find( "patch") != -1 ):
                dateTime = result[ idx1 : idx2 ]
                versions.append( dateTime )
                
        
        if( len(versions) == 0  ):
            print "no previous record"
            return []

        # sort
        versions.sort( reverse=True )

        return versions


    def get_versions_short( self, fName, dateTime, toPrint=True ):
        '''  extract the subset of patch files, which can 
            recover fName to version of dateTime
        '''

        versions = self.get_file_versions( fName )  
         
        # check the input
        inputCorrect=False
        for version in versions:
            if dateTime == version:
                inputCorrect = True

        if( inputCorrect == False ): 
            print "%s is not available" %( dateTime )
            return []
        else:
            if( toPrint ):
                for ii in range( len(versions) ):
                     
                    if( dateTime == versions[ii] ):
                        pass
                        print "%s <== selected" %( dateTime )
                    else:
                        print versions[ii]
         

        # get the shortened versions
        versions_short = []
        toBreakNextTime = False
        for version in versions:
            if( dateTime != version ):
                versions_short.append( version )
            elif( dateTime == version ):
                toBreakNextTime = True

            if( toBreakNextTime ): break

        return versions_short
        

    def back_to_previous_version( self, fName, dateTime ):
        # todo: remove debugFlag. we will create testing function
        # somewhere else.

        debugFlag = False

        versions_short = self.get_versions_short( fName, dateTime, False )
        
        if( len(versions_short) == 0 ): return

  
        
        print "Current: "
        do_cat( fName )   
         
        
        # apply patches   
        patchName = "" 
        for version in versions_short:
            
            patchName = self.backupDir + "/" + self.historyDir + "/" \
                + version + "/patch/" +  fName + ".patch"

            if( debugFlag ):     
                print "-----------------------"   # for debug
                print "patch version = ", version # for debug 
             
                     
            applyPatch_Reverse( fName, patchName )
            
            if( debugFlag ): 
                print "after apply %s" %version  # for debug
                do_cat( fName )                  # for debug

        print "roll back to %s: " % self.convert_to_readable( version )         
        do_touch( dateTime, fName ) # change the modification time.  
        do_cat( fName )
        pass



    def get_dateTime_from_pathfile( self, pathFile ):
        ''' currently, it is not in used.
            return is a string ex. "201901010005"
        '''
        with open( pathFile, "r" ) as f:
            line = f.readline()
            line = line.strip()
        items = line.split()
        year_month_day = items[2]
        hh_mm = items[3]
        hh_mm = hh_mm[:5]
        year_month_day = year_month_day.split("-")
        hh_mm = hh_mm.split(":")
        outStr = ""
        for xx in year_month_day: outStr += xx
        for xx in hh_mm: outStr += xx    
        
        return outStr 

         

    def view_previous_version( self, fName, dateTime ):

        debugFlag = False
        
        versions_short = self.get_versions_short( fName, dateTime, True )

        
        # clean the old temp working folder (seems not necessary)    
        # dirToRemove = self.backupDir + "/" + self.tempDir 
        # remove_a_folder( dirToRemove )

        # make a temp working folder
        tempDir = self.backupDir + "/" + self.tempDir 
        make_a_fold( tempDir )

        # make a copy  
        do_rsync( fName, tempDir )
       
        # deal with file name
        fNameTemp = get_fileName( fName )
        fNameTemp = self.backupDir + "/" + self.tempDir + "/" + fNameTemp
        
        if( debugFlag ):
            print "-----------------------"   
            print "latest version"            
            do_cat( fNameTemp )               

        # apply patches        
        for version in versions_short:
            
            patchName = self.backupDir + "/" + self.historyDir + "/" \
                + version + "/patch/" +  fName + ".patch"

            if( debugFlag ):    
                print "-----------------------"  # for debug
                print "patch version = ", version
                print "debug fNameTemp = ", fNameTemp
                    
            applyPatch_Reverse( fNameTemp, patchName )
            
            if( debugFlag ):    
                print "after apply %s" %version 
                do_cat( fNameTemp ) #for debug

        # print on the screen.    
        do_cat( fNameTemp )    
        pass    
         
    ###################################################
    # methods related to deleted files.

    def show_all_deleted_files( self ):
        ''' print out all deleted files in backup/history/version/deleted/ '''
        historyDir = self.backupDir + "/" + self.historyDir + "/"
        results = get_find_results( historyDir, "-type f" )
         
        for fName in results:
            if( fName.find("deleted/") != -1 ):
                fName = fName.replace( historyDir ,"" )
                fName = fName.replace( "/deleted/" ," " )
                print fName
        pass        

     


    ###################################################
    # methods related to recover a dir to a dateTime.

    def recover_a_dir( self, dirName, dateTime ):
        pass    
    

    ###################################################
    # methods for history versions

    def get_run_dry_result( self , debug = False ):
        ''' return is a list, each item is a line for the change record'''
        
        if( debug ):
            print "inside get_run_dry_result "

        for source in self.sources:
            dry_run_results = \
                do_rsync_dry_run( source, self.backupDir )
            
            if( debug ):
                print "source:", source

            for item in dry_run_results:
                if (item not in self.dry_run_results and \
                    len(item) != 0 ):
                    self.dry_run_results.append( item )
                    if( debug ):
                        print "dry_run_result:", item
        
        if( debug ):
            print "end\n"

        return self.dry_run_results

    def convert_to_machine( self, version ):
        return version.replace("_","" )    
        
    def convert_to_readable( self, version ):
        return version[:4] + "_" + version[4:8] + "_" + version[8:]

    def get_history_versions( self ):
        ''' get all the dirnames (dateTime) under history dir''' 
        historyDir = self.backupDir + "/" + self.historyDir + "/"
        self.history_versions = do_ls( historyDir ) 
        return self.history_versions

    def manage_history_version( self, debug = False ):
        
        self.get_history_versions()
        
        today = date.today()

        if( debug ):
            print "inside manage_history_version()"
            print "the following folders are to be removed:"

        for version in self.history_versions:
            year   = int( version[  :4] )
            month  = int( version[ 4:6] ) 
            day    = int( version[ 6:8] )
            # hour   = int( version[ 8:10] )
            # minute = int( version[ 10:12] )

            version_date = date( year, month, day )
            deltaTime = abs( today - version_date )
            deltaTime_days = deltaTime.days

            if ( self.upperDays < deltaTime_days ):
                dirName = self.backupDir + "/" + self.historyDir + "/" + version
                remove_a_folder( dirName )
                if( debug ):
                    print "remove folder: %s" %(dirName)
      
            pass




__doc__= """RSYNC_BACKUP  

Usage:  script.py [-r BACKUP_PLAN     ] 
                  [-b MAIN_BACKUP_DIR ]   
                  [-hvi]   

Options:
    -h --help                Show this screen
    -v --version             Show version.
    -r --read BACKUP_PLAN    Read in a backup plan, and process it. 
    -i --interactive
    
    example
    # process a backup plan
    ./script -r backup_plan

    # process a backup plan, and then go in interaction menu.
    ./script -ir backup_plan    

    

"""

if __name__ == '__main__':
    
    rsync_tool = RSYNC()

    

    arguments = docopt(__doc__, version='RSYNC_BACKUP1.0' )
    # print("\n this is the arumgents from docopt:\n",  arguments )
    
    # process a backup plan
    if( arguments['--read'] and arguments['--interactive'] ):
        # not finish yet.
        rsync_tool.read_in_backup_plan( arguments['--read'] )
        rsync_tool.process()
        rsync_tool.show_menu()
        
    elif( arguments['--read'] ):
        rsync_tool.read_in_backup_plan( arguments['--read'] )
        rsync_tool.process()

    
    
 