#include <iomanip>
#include <iostream>
#include <string>
#include <memory> // to use shared pointer.
using namespace std;


/*
   shared pointer

   memory leak: 
   when we create a pointer to an object, ex.

    { 
	Dog* ptr = new Dog(); 
	// do sth...
    }
    but we forget to 'delete ptr' when it out of the scope, 
    which will cause memory leak.

*/

// a class to demo the shared pointer.
class Dog 
{
    private:
	string _name;
    public:
    Dog(){ _name ="nameless";}      // default constructor. Dog dog1():
    Dog( string name ){ _name = name; }


    Dog( const Dog&  rhs);           // copy constructor. Dog dog2( dog1 ); 
    Dog& operator=(const Dog& rhs);  // copy assignment operator. dog3 = dog2;
    ~Dog(){};

    Dog( Dog&& );             // move constructor  Dog dog4( create_dog() );
    Dog& operator=( Dog&& );  // move assignment operator. Dog dog5 = create_dog(); 

    void bark() {
    
	    cout << "Dog " << _name << " barks." << endl;
    }
};


void func_demo() 
{
    
    shared_ptr<Dog> ptr1 = make_shared<Dog>( "Tom" ); 
    // preferred way.
    // make_shared<class_name> ( args for constructor );
    // it provides some safety checks.
    // it makes sure ptr1 and Dog("Tom") are created successfully.

    // shared_ptr<Dog> ptr1( new Dog("Tom") ); // this one also works.

    /*
     *
     * NOT a good way:
     * Dog* d = new Dog("Tank");
     * shared_ptr<Dog> p(d);  // p.use_count() == 1
     * shared_ptr<Dog> p2(d); // p2.use_count() == 1
     * When we go out of scope p2 will be destroyed first,
     * the 'd' with Dog("Tank") will be destroyed.
     * and then when p is destroyed,
     * then 'd' is destroyed again ( but it is not there anymore)
     * It will cause undefined behavior.
     *
     * an object should be assigned to a smart pointer as soon as
     * it is created, and the 'raw' pointer should not be used again.
     */

    ptr1->bark(); 
    // use '->' to access Dog class.

    cout << ptr1.use_count() << endl;
    // use '.' to access shared pointer. it will print 1

    shared_ptr<Dog> ptr2 = ptr1;
    cout << ptr1.use_count() << endl; // it will print 2
    cout << ptr2.use_count() << endl; // it will print 2


    (*ptr2).bark();
    // de-reference.

}
// when smart pointers are out of scope, they are deleted.
// so we won't have the issue of memory leak.


void func_demo2 () 
{

    shared_ptr<Dog> ptr1 = make_shared<Dog>("Gunner");
    shared_ptr<Dog> ptr2 = make_shared<Dog>("Tank");

    ptr1 = ptr2; 
    // at this point, Dog("Gunner") will be deleted.
    // since no smart pointer pointing to Dog("Gunner")
    // ptr1 now is pointing to "Tank" as ptr2.
    //
    // if we do:
    // ptr1 = nullptr; // or 
    // ptr1.reset();
    // then Dog("Gunner") is also deleted.



    // to make a costumed deleter.
    // by default, the deleter is the 'operator delete'
    // we use a costumed deleter, by passing a lambda function as following.
    shared_ptr<Dog> ptr3 
	= shared_ptr<Dog>( new Dog("Mook"), 
	    [] (Dog* p) { cout <<"custom deleting" << endl; delete p; }  );



    shared_ptr<Dog> ptr4( new Dog[3], [] (Dog *p) { delete [] p; });
    // ptr4 is only pointing to the first Dog[0], 
    // and Dog[1] and Dog[2] will have memory leak.
    // so we have use 'delete []' to delete whole Dog objects.


}

/*
 * avoid to use raw pointer.
 */
void func_demo3 () 
{
    shared_ptr<Dog> ptr1 = make_shared<Dog>("Tom");

    // avoid to use 'raw' pointer !!
    Dog* d = p1.get(); // get() will return 'raw' pointer.

}




int main() {

    func_demo();

    func_demo2();
}
