#include <iomanip>
#include <iostream>
#include <string>
#include <memory> // to use shared pointer.
using namespace std;


/*
    weaker pointer

    solution for the issue of Cyclic reference.

    shared_ptr has ownership of the object.

    weak_ptr doesn't have the ownership of the object, 
    ( when and how the object is deleted is not the weak_ptr's business)
    but it can access the object. 

    weak_ptr is similar to raw pointer, but it provides one level protection
    that no operation can perform delete on the weak_ptr. 

*/

// a class to demo the shared pointer.
class Dog 
{
    private:
	string _name;
	//shared_ptr<Dog> my_Friend; ( this will cause Cyclic reference) 
	weak_ptr<Dog> my_Friend;
    public:
    Dog(){ _name ="nameless";}      // default constructor. Dog dog1():
    Dog( string name ){ _name = name; }


    Dog( const Dog&  rhs);          
    Dog& operator=(const Dog& rhs);  
    ~Dog(){};

    Dog( Dog&& );            
    Dog& operator=( Dog&& ); 

    void bark() 
    {
	    cout << "Dog " << _name << " barks." << endl;
    }

    void makeFriend( shared_ptr<Dog> p ) { my_Friend = p; }

    void showFriend( ) {
	
	if( !my_Friend.expired() ) 
	{
	    cout << "my friend is: " << my_Friend.lock()->_name  << endl;
	}
	// to use weak_ptr, we have to use lock()
	// it creates a shared_ptr out of the weak_ptr,
	// to make sure the weak_ptr is still pointing to a valid object.
	// and make sure the object will not be deleted.

	// the weak_ptr.lock() will throw an exception,
	// when the my_Friend is empty. 
	// so we can add a check, by
	// !my_Friend.expired()
    }
};


void func_demo() 
{

    shared_ptr<Dog> ptr1 ( new Dog("Gunner") );
    shared_ptr<Dog> ptr2 ( new Dog("Smokey") );

    ptr1->makeFriend( ptr2 );
    ptr2->makeFriend( ptr1 );
    // here, we will make ptr2 <--->ptr1 ( cyclic reference )
    // they won't be deleted when they are out of scope.

    ptr1->showFriend();
}





int main() {

    func_demo();

}
