#include <iostream>
#include <vector>
#include <deque>
#include <list>
#include <set> // with set and multiset class
#include <map> // with map and multimap class
#include <unordered_set> // with unordered_set/multiset
#include <unordered_map> // with unordered_map/multimap
#include <algorithm> 
#include <numeric>      // some numeric algorithm
#include <functional>   // functor.

using namespace std; // STL library is in the std namespace.

/*
    containers:
    * Sequence containers ( array and lined listed )
    c++ provides: 
    vector,
    deque,
    list,
    forward_list,
    array.
    
    * Associative containers ( implemented in binary tree, always sorted )
    c++ provides:
    set, multiset,
    map, multimap.

    * Unorder containers ( implemented by hash table )
    c++ provides:
    Unordered set/ multiset,
    Unordered map/ multimap.
*/



int main() {

    // Sequence container.
    vector<int> vec;

    vec.push_back( 4 );
    vec.push_back( 1 );
    vec.push_back( 8 );

    vec.size();  // the method size() return the numbers of elements in the container.

    // vector element can be randomly access, it means
    // we can do the following.
    cout << vec[2] << endl;;   // no range check.
    cout << vec.at(2)<< endl; // having range check, throw range_error exception.:w



    // because element in vector class can be random access
    // we can access it like array.
    // for( int i=0; i<vec.size(); i++) { cout << vec[i] << endl; }
    // but this is not recommended.
    // the better way is to use iterator.
    // it is faster than the random-access method.
   
    // looping in C++11

    for( auto itr: vec) {
		cout << "item = " <<  itr << endl;
    }



    // pointer to vector

    int* ptr = & vec[0]; 
    ptr[2] = 6;
    // let a pointer to store the address of the first element in the vector
    // 'vec', then we can access it like an array.

    for( auto itr: vec) {
		cout << "new item = " <<  itr << endl;
    }



    //===============================================================
    // common API
    //
    
    if( vec.empty() ) { cout << "no elements in vec" << endl; }

    vector<int> vec2( vec ); 
    // invoke the copy constructor of the vector class
    // so that we copy the element in vec to vec2.

    vec.clear(); // clear all the elements in 'vec'.

    vec2.swap( vec ); 
    //  vec2 <===> vec, 
    //  so vec2 is empty, and vec has original vect2 data.


    //=================================================================
    // property of vector:
    //  since it only grows only  to the end.
    //  Fast: insert and remove at the end: O(1)
    //  Slow: insert and remove at the begin and middle: O(n)
    //        you have to shift all the items.
    //  Slow: Search: O(n)




    //=================================================================
    // property of deque:
    // it can grows at the end and at the beginning.
    //  Fast: insert and remove at the end and beginning: O(1)
    //  Slow: insert and remove in the middle: O(n)
    //        you have to shift all the items.
    //  Slow: Search: O(n)

    deque<int> deq = { 4, 6, 7};
    deq.push_front(2);   // deg: { 2, 4, 6, 7 }
    deq.push_back(3);    // deg: { 2, 4, 6, 7, 3 }

    // very similar property to vector.
    // it has random access. 
    // ex. cout << deq[0] << endl;



    //=================================================================
    // property of list:
    // it is doubly-linked list, each element has pointers, which pointing to
    // the next and previous elements.
    //
    //  Fast: insert and remove at anywhere: O(1)
    //  Slow: Search: O(n), much slower than vector.
    //                vector store data in continuous way in memory.
    //                but list store data in different memory place.
    //  NO random access ( no [] operator )
    //  killing function: splice

    list<int> mylist = { 5, 2, 9};
    mylist.push_back(6);   // { 5, 2, 9, 6}
    mylist.push_front(4);  // { 4, 5, 2, 9, 6}

    auto itr = find( mylist.begin(), mylist.end(), 2 );
    // find the element = 2.
    // itr is pointing the third element.

    // we insert an element 8, in front of element with value = 2.
    mylist.insert( itr, 8 );  
    // { 4, 5, 2, 9, 6}    before.
    // { 4, 5, 8, 2, 9, 6} after insert. 
    // list is faster than vector and deque.

    itr++; 
    // increate the iterator,
    // now it points to the elment with value '9' from '2'.

    mylist.erase( itr ); 
    // erase the element.
    // { 4, 5, 8, 2, 6} after erase. 

    cout << "list item: " ;
    for( auto item: mylist ) { cout << item  << " "; } 
    cout << "\n"; 

    list<int> mylist2 = { 1, 2, 3 };
    // auto itra = find( mylist2.begin(), mylist2.end(), 1);
    // auto itrb = find( mylist2.begin(), mylist2.end(), 3);
    auto itra =  mylist2.begin();
    auto itrb =  mylist2.end();

    itr = find( mylist.begin(), mylist.end(), 5 );
    mylist.splice( itr, mylist2, itra, itrb );
    // what splice is doing is to cut the some range of data in 
    // mylist2, then insert these data infront of itr position of mylist.
    // list item: 4 5 8 2 6 
    // list item: 4 1 2 3 5 8 2 6 
    
    cout << "list item: ";
    for( auto item: mylist ) { cout << item  << " "; } 
    cout << "\n"; 

    
    //=================================================================
    // property of forward_list:
    // it is single-linked list, each element has a pointer to the next elnt.




    //=================================================================
    // property of array:
    // array container is a very thin layer around the native array.
    
    // conisder an array.
    int arr[3] = { 3, 4, 5};
    // you cannot do arr.begin(), end(), size(), swap()...
    // the solution is:

    array<int, 3> a = { 3, 4, 5 };
    // use meta class array, and put the parameter 'int' and 3
    // it is represent an integer array of 3 elements.

    cout << "size of array<int,3> = " <<  a.size() << endl;

    // note: 
    // array<int, 3> will have different type than
    // array<int, 4>
    
    // the size of array container cannot be changed.

}
