#include <iostream>
#include <string>
#include <vector>
#include <deque>
#include <list>
#include <set> // with set and multiset class
#include <map> // with map and multimap class
#include <unordered_set> // with unordered_set/multiset
#include <unordered_map> // with unordered_map/multimap
#include <algorithm> 
#include <numeric>      // some numeric algorithm
#include <functional>   // functor.

using namespace std; // STL library is in the std namespace.

/*
    
    
    //=================================================================
    // property of unordered associative container (c++11)
    // unordered set, multiset,
    // unordered map, multimap.
    it uses a hash table ( or called hash function )
    unordered associative container is an array and linked lists.

    the array is also called the "array of buckets".
    in each element, there is a linked list.

    each elment is calculated by hash function, and it gets a value.
    according to the value, the element is inserted to the corresponding list.
    
    if we have an efficient hash function, then find the elment in the list is 
    very fast. It only takes constant time.

    // Fastest search/insert at any place: O(1)
    //	    reminder: Associative containers take O( log(n) )
    //	    reminder: vector, deque containers take O( n )
    //	    reminder: list container takes O(1) to insert and  O( n ) to search
    // Unorder set/multiset: element value cannot be changed.
    // Unorder map/multimap: element key   cannot be changed.
*/



int main() {


    // demo: unordered_set 

    unordered_set<string> mySet = { "red", "green", "blue" };


    // the search is only constant time, O(1)
    // find by value for 'set'
    unordered_set<string>::const_iterator itr = mySet.find( "green" );
   

    // for inserting. It is important to check.
    // we have to set not duplicated elements.
    itr = mySet.find( "yellow" );
    if( itr == mySet.end() ) { mySet.insert( "yellow" );  }



    // insert a set of data from a vector container.
    vector<string> vec = { "purple", "pink" };
    mySet.insert( vec.begin(), vec.end() );

    cout << "TEST insert data from a vector " << endl; 
    for( auto item : mySet ) { cout << item << " "; } cout << "\n";




    //==============================================================

    // hash table specific APIs:

    cout << "load_factor = " << mySet.load_factor() << endl;
    // load factor = the ratio between total number of elements and 
    // total number of buckets.
    
    string x = "red";
    cout <<  x << " is in bucket #" << mySet.bucket( x ) << endl;


    cout << "Total bucket # =  " << mySet.bucket_count() << endl;
    // note: we store 6 elements in the set.
    // and currently, we have 7 buckets. 
    // the load factor = 6/7 = 0.8571...

    // unordered_set: a set using hash table, no duplicated elements.
    // unordered_multiset: unordered set that allows duplicated elements.
    // unordered_map: unordered set of pairs.
    // unordered_multimap: unordered map which allows to have duplicated keys.



    // hash collision: 
    // too many items are inserted at the same bucket.
    // so the performance of searching is degraded to O(n).
    //   reminder: associated container guarantee O( log(n) ).



	
    //=========================================================

    // demo: unordered_map

    unordered_map<char, string> day 
	= { {'S', "Sunday"}, 
	    {'M', "Monday"} };

    cout << day['S'] << endl; 
    cout << day['T'] << endl;  // no range check, no error will appear.
    cout << day.at('M') << endl; // having range check recommended way.


    // insert 
    day['W'] = "Wednesday"; 
    // it is the same as  day.insert( make_pair( 'W', "Wednesday" ) );
    day.insert( make_pair( 'F', "Friday" ) );


    // to modify
    day['W'] = "Wednesday!!";
    // when we use [ ] 
    // the compiler will assume we want to modify 'day'
    // even when we just want to print out.
    // like cout << day['W'] << endl;




    //==========================================================

    /*
     * another way to categorizing containers:
     *
     * 1. array based containers: vector, deque.
     *    -- it will invalidate the pointer.
     *
     * 2. node based container: list, associated containers, unordered
     * containers.
     *    -- it will NOT invalidate a pointer.
     */

    vector<int> aVec = { 0, 1, 2, 3} ;

    int* p = & aVec[2]; // point to  '2'

    aVec.insert( aVec.begin(), -1 ); // { -1, 0, 1, 2, 3 }

    // now p will be invalidated .. it can not be used anymore.





    //============================================================

    /*
	container adaptors
	they provide a restricted interface to meet speical needs.	

	1. stack: LIFO, push(), pop(), top()

	2. queue: FIFO, push(), pop(), front(), back()

	3. priority queue: first item always has the greatest priority
			push(), pop(), top()
    */



}
