#include <iostream>
#include <chrono>  // for high_resolution_clock


// ==============Container===================
// ref: https://en.cppreference.com/w/cpp/container

#include <vector>
// Random access - constant O(1)
// Insertion or removal of elements at the end - amortized constant O(1)
// Insertion or removal of elements - linear in the distance to the end of the vector O(n) 

#include <deque>
// Random access - constant O(1)
// Insertion or removal of elements at the end or beginning - constant O(1)
// Insertion or removal of elements - linear O(n) 

#include <list>
#include <forward_list>

//=================== Container adaptors ======================
#include <stack>  // last-in-first-out
#include <queue>  // first-in-first-out

using namespace std;

void speedTest_vector();
void speedTest_deque();
void speedTest_list();
void speedTest_forward_list();
 
void speedTest_stack();
void speedTest_queue();

int main() {
   
   speedTest_stack();
    
   // speedTest_vector();
   speedTest_deque();


   
   // speedTest_queue();  

   // speedTest_list();
   // speedTest_forward_list();

    return 0;
}



void speedTest_vector(){

    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    vector<int> myVector;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myVector.push_back( 10 );
        myVector.push_back( 11 );
        myVector.push_back( 12 );
        
        myVector.clear();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();
    
    std::chrono::duration<double> elapsed = finish - start;
    
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "vector" );
}

void speedTest_deque(){

    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    deque<int> myDeque;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myDeque.push_back( 10 );
        myDeque.push_back( 11 );
        myDeque.push_back( 12 );
        
        myDeque.clear();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();
    
    std::chrono::duration<double> elapsed = finish - start;
    
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "deque" );
}

void speedTest_list(){

  // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    list<int> myList;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myList.push_back( 10 );
        myList.push_back( 11 );
        myList.push_back( 12 );
        
        myList.clear();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();
    
    std::chrono::duration<double> elapsed = finish - start;
    
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "list" );

}

void speedTest_forward_list(){

    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    forward_list<int> myForward_list;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myForward_list.push_front( 10 );
        myForward_list.push_front( 11 );
        myForward_list.push_front( 12 );
        
        myForward_list.clear();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();
    
    std::chrono::duration<double> elapsed = finish - start;
    
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "forward_list" );
}

void speedTest_stack() {
    
    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    stack<int> myStack;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myStack.push( 10 );
        myStack.push( 11 );
        myStack.push( 12 );
        
        myStack.pop();
        myStack.pop();
        myStack.pop();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();

    std::chrono::duration<double> elapsed = finish - start;

    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "stack" );
}
 

void speedTest_queue() {

    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    queue<int> myQueue;
    
    for( int ii = 0; ii < 10000000; ii++ ){
    
        myQueue.push( 10 );
        myQueue.push( 11 );
        myQueue.push( 12 );
        
        myQueue.pop();
        myQueue.pop();
        myQueue.pop();
    }


    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();

    std::chrono::duration<double> elapsed = finish - start;

    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "queue" );
    // std::cout << "Elapsed time: " << elapsed.count() << " s (queue) \n";
}