#include <iostream>


#include "spdlog/spdlog.h" // require C++ 11
#include "spdlog/sinks/basic_file_sink.h"
#include "spdlog/sinks/stdout_color_sinks.h" // console logger


#include <memory> // for make_shared()
using std::cout; 
using std::endl;

// ref: http://www.programmersought.com/article/2845994638/

// the very basic spdlog
void demo1(){

    // Use the default logger (stdout, multi-threaded, colored)
    // It provides a python like formatting API using the bundled fmt lib
    spdlog::info("Welcome to spdlog!");
    spdlog::error("Some error message with arg: {}", 1);
    
    spdlog::warn("Easy padding in numbers like {:08d}", 12);
    spdlog::critical("Support for int: {0:d};  hex: {0:x};  oct: {0:o}; bin: {0:b}", 42);
    spdlog::info("Support for floats {:03.2f}", 1.23456);
    spdlog::info("Positional args are {1} {0}..", "too", "supported");
    spdlog::info("{:<30}", "left aligned");
    
    spdlog::set_level(spdlog::level::debug); // Set global log level to debug
    spdlog::debug("This message should be displayed..");    
    
    // change log pattern
    spdlog::set_pattern("[%H:%M:%S %z] [%n] [%^-%L-%$] [thread %t] %v");
    
    // Compile time log levels
    // define SPDLOG_ACTIVE_LEVEL to desired level
    SPDLOG_TRACE("Some trace message with param {}", {});
    SPDLOG_DEBUG("Some debug message");
    
    // Set the default logger to file logger
    auto file_logger = spdlog::basic_logger_mt("basic_logger", "logs/basic.txt");
    spdlog::set_default_logger(file_logger); 
	
    spdlog::info("tai message!") ;  // now the info is written to file.      
}

// console logger
void demo2()
{
    spdlog::set_pattern("[%H:%M %Y-%m-%d] [%n] [%^%L%$] [th %t] %v");
    cout << "console logger will print info on the screen " << endl;

    auto logger_console = spdlog::stdout_color_mt("console"); // name = console

    // the "name" can be used to retrieve the logger

    spdlog::get("console")->info( "this is from retrieval." );
    spdlog::get("console")->critical("....oh no ....");
}

// file logger
void demo3()
{
    spdlog::set_pattern("[%H:%M %Y-%m-%d] [%n] [%^%L%$] [th %t] %v");
    cout << "file logger will write to a file, see logs/log.txt " << endl;

    bool truncate = true;
    auto file_logger= spdlog::basic_logger_mt("file", "logs/log.txt", truncate ); 
    // name = file
    // if truncate = true, it remove the old content.

    // the "name" can be used to retrieve the logger
    spdlog::get("file")->info( "writing into file1." );
    spdlog::get("file")->critical("writing into file2.");
}


/*
    suppose we want to save information in two places. 
    1. console, the level is "warn".
    2. file, the level is "trace".

    here, we create two sinks
    1. for console_sink from spdlog::sinks::stdout_color_sink_mt
    2. for file_sink from spdlog::sinks::basic_file_sink_mt

    then we create a logger connect these two sinks.


*/
void demo4(){

    auto console_sink = std::make_shared< spdlog::sinks::stdout_color_sink_mt >();

    // tunning console sink.
    console_sink->set_level(spdlog::level::warn);
    // console_sink->set_pattern("[multi_sink_example] [%^%l%$] %v");

    auto file_sink = std::make_shared< spdlog::sinks::basic_file_sink_mt >("logs/multisink.txt", true);
    
    // tunning file sink.
    file_sink->set_level(spdlog::level::trace);

    spdlog::logger logger( "multi_sink", {console_sink, file_sink} );
    
    logger.set_level( spdlog::level::debug );

    logger.warn( "this should appear in both console and file");
    
    logger.info( "this message should not appear in the console, only in the file" );

}

/*
    it is similar to demo4, but
    here we set the "multi_sink" logger as the default logger,

    and use try...catch block.
*/
void demo5(){

    try
    {
        // console sink and tunning.
        auto console_sink = std::make_shared<spdlog::sinks::stdout_color_sink_mt>();
        console_sink->set_level(spdlog::level::warn);
        console_sink->set_pattern("[multi_sink_example] [%^%l%$] %v");

        // file sink and tunning
        auto file_sink = std::make_shared<spdlog::sinks::basic_file_sink_mt>("logs/multisink.txt", true);
        file_sink->set_level(spdlog::level::trace);

        spdlog::logger logger("multi_sink", {console_sink, file_sink});
        logger.set_level(spdlog::level::debug);
        logger.warn("this should appear in both console and file");
        logger.info("this message should not appear in the console, only in the file");

        // set multi_sink logger as default logger
        spdlog::set_default_logger(std::make_shared<spdlog::logger>("multi_sink", spdlog::sinks_init_list({console_sink, file_sink})));

    }
    catch (const spdlog::spdlog_ex& ex) // ex for exception.
    {
        std::cout << "Log initialization failed: " << ex.what() << std::endl;
    }


}




int main()
{
    
    // demo1();
    // for( int i=0;i<80;i++){ cout <<"-"; } cout <<"\n";

    // demo2();
    // for( int i=0;i<80;i++){ cout <<"-"; } cout <<"\n";

    // demo3();
    // for( int i=0;i<80;i++){ cout <<"-"; } cout <<"\n";


    // demo4();

    demo5();

    return 0;
}
