/*
    basic property for multi-threading (MT)
    Pros:
    -- faster to start than multi-process (MP)
    -- lower overhead ( few system protection needed than MP )
    -- communicate via shared memory 

    Cons:
    -- MT is difficult to implement than MP
    -- MT cannot run on a distributed system.

********************************************************************/

#include <thread> // to compile it we need g++ with -pthread
#include <iostream>
 

void function1( ){

    std::cout << "from function1 " << std::endl;
    /*
        this function1 is for demo, but actually,
        it is not a good design,
        since we can race for "std::cout" from main and child threads.
    */
}


/* 
    once entering the main(), we start the main thread.
    by calling std::thread t1(...)
    we create a child thread t1, and t1 starts its job.

    we can either detach/join the t1 thread to main thread.
    -- join means main thread wait for t1 to finish its jobs.
    -- detach means let t1 thread run freely, when the main finish 
       (thus the program terminates), t1 may not finish its jobs yet. 

    if we don't have either t1.join() or t1.detach()
    the system will print out 
    "terminate called without an active exception"

    we can do join() or detach() just once.


    how many physical threads we can have?
    we can use std::thread::hardware_concurrency() to get this info.
    once our threads over this number, it calls "over-subscription."
*/
int main(){

    std::cout << "thread ID for main  is " <<  std::this_thread::get_id()  
    << ". " << std::endl;

    std::thread t1( function1 );

    auto threadID = t1.get_id();
    std::cout << "thread ID for t1    is " <<  threadID  << ". " << std::endl;

    int NTHREAD = std::thread::hardware_concurrency() ;
    printf("we have %d thread(s) in this hardware.\n", NTHREAD  );



    t1.join();  // let main thread wait for t1 to finish.
    //t1.detach(); // let t1 run separately.

    // since we have performed join() or detach()
    // do t1.join() again will make compilation fail.
    if( t1.joinable() ){
        t1.join();
    }
    


    return 0;
}


/*
	http://www.cplusplus.com/reference/thread/thread/thread/
	constructor (initialization) syntax:
	
	template <class Fn, class... Args>
	explicit thread (Fn&& fn, Args&&... args);
*/
