/*
   in this demo, we show the concept of deadlock.
   we will also discuss how to avoid deadlock. 

   note: this program will be halted ( deadlock ).

   Why deadlock:
   -- we have two std::mutex object ( m_mutex1 and m_mutex2 )
   -- m_mutex1 is waiting for m_mutex2, also 
      m_mutex2 is waiting for m_mutex1.  

	
   how to avoid deadlock

   1. use single lock_guard at a time ( no multiple ones )
   2. if we have to use multiple lock_guard, set them in the same order.
   3. use std::lock()
   e.g.

   std::lock( m_mutex1, m_mutex2 ); 
   std::lock_guard< std::mutex> guard1( m_mutex1, std::adopt_lock )
   std::lock_guard< std::mutex> guard2( m_mutex2, std::adopt_lock )

********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>    // std::call_once, std::once_flag 
#include <iostream>

using namespace std;


/*
 * note: we use "cout" for demo purpose only.
 */
std::once_flag  _onceFlag;


class LogFile{
  private:
  
    std::mutex m_mutex1; 
    std::mutex m_mutex2; 
  
  public:
    LogFile() { } 

    ~LogFile() { }

   
	/*
	 * call by t1 thread first, m_mutex1 is locked.
	 * but the main thread, also set m_mutex2 locked.
	 * and then in shared_printA(), it cannot go beyond guard2( m_mutex2 )
	 * therefore a deadlock happens.
	 */

    void shared_printA( int threadID, int iValue) {

        std::lock_guard< std::mutex>  guard1( m_mutex1 );
        std::lock_guard< std::mutex>  guard2( m_mutex2 );

		// solution:
		// std::lock( m_mutex1, m_mutex2 ); 
		// std::lock_guard< std::mutex> guard1( m_mutex1, std::adopt_lock )
		// std::lock_guard< std::mutex> guard2( m_mutex2, std::adopt_lock )

        cout << "A -- I am from " << threadID << " thead: "<< iValue << endl;
    }
    
	void shared_printB( int threadID) {

        std::lock_guard< std::mutex>  guard2( m_mutex2 );
        std::lock_guard< std::mutex>  guard1( m_mutex1 );
		
		// solution:
		// std::lock( m_mutex1, m_mutex2 ); 
		// std::lock_guard< std::mutex> guard1( m_mutex1, std::adopt_lock )
		// std::lock_guard< std::mutex> guard2( m_mutex2, std::adopt_lock )

        cout << "B -- I am from " << threadID << endl;
    }
};




// the enter function called by thread.
void Printing_job( LogFile& log, int threadID ){
    
    for( int i=0; i<100; i++) {
        log.shared_printA( threadID , i );    
    }
    
}
 


 
int main(){

    LogFile log;

    thread t1( Printing_job, std::ref(log), 1 );
    
	// call by the main thread.
    for( int i=0; i<100; i++) {
        log.shared_printB( 0  );    
    }
 
    t1.join();

   
    return 0;
}
