/*
    Lazy Initialization ( initialization upon first use idiom )

    if( ! f.is_open() ){
        std::lock_guard< std::mutex>  guard( m_mutex_open  );
        f.open("log2.txt", ios::out ); f.close();
        f.open("log2.txt", ios::out );  
    }

    this way is not thread safe.

    threadA goes first, it sees f is not open, then enter the if block.
    it sees the lock_guard statement, then lock the process.

    threadB starts, it also sees f is not open, then enter the if block,
    but since threadA already make the guard locked, and threadB is waiting.

    ok, now threadA finish the process, and release m_mutex_open.
    threadB sees the statement of f.open(..), and so it opens the file.
    ... so we can end up open files by different threads.

    the correct way to fix it is:

    std::lock_guard< std::mutex>  guard( m_mutex_open  );
    if( ! f.is_open() ){    
        f.open("log2.txt", ios::out ); f.close();
        f.open("log2.txt", ios::out );  
    }

	or one can use call_once.
********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>    // std::call_once, std::once_flag 
#include <iostream>
#include <fstream>

using namespace std;



std::once_flag  _onceFlag;


class LogFile{
  private:
  
    std::mutex m_mutex; 
    std::mutex ; 
  
    fstream f;  // we should not let outside to access f
  
  public:
    LogFile() { }

    ~LogFile() { f.close(); }

   

    void shared_print( int threadID, int iValue) {

        // m_mutex_open is used for open and clean the file, 
        // which will should be use for once.
        // notice: I move the same piece of code from constructor to here.
        // actually, it can be replaced by using std::call_once

        std::lock_guard< std::mutex>  guard( m_mutex_open  );
        if( ! f.is_open() ){
            f.open("log2.txt", ios::out ); f.close();
            f.open("log2.txt", ios::out );  
        }
           
        
        // m_mutex is for locking the fstream write out,
        // which will be used several times.
        std::lock_guard< std::mutex>  guard( m_mutex ); 
        f << "I am from " << threadID << " thead: "<< iValue << endl;
        
    }
};




// the enter function called by thread.
void Printing_job( LogFile& log, int threadID ){
    
    for( int i=0; i<5; i++) {
        log.shared_print( threadID , i );    
    }
    
}
 


 
int main(){

    LogFile log;

    thread t1( Printing_job, std::ref(log), 1 );
    thread t2( Printing_job, std::ref(log), 2 );
     
    // call by the main thread.
    for( int i=0; i<5; i++) {
        log.shared_print( 0, 0  );    
    }
 
 
    t1.join();
    t2.join();
    cout << "see log2.txt" << endl;

   
    


    return 0;
}
