/*
    demo: condition_variable class, wait(), notify_one(). 

    suppose we have two threads, threadA and threadB.

    threadA is responsible for generating data and filling it in data container.
    
    threadB is to exam whether data container is empty or not, if not empty,
    then read its value. we want to do it in real time.
    
    normally, we can set threadB to check the data container and then sleep 
    for a fixed period of time. But this is not easy to set the sleep duration.

    one of the solution is to let threadA to send a notification to wake up 
    threadB whenever the data is generated. 

********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>    // std::call_once, std::once_flag 
#include <iostream>
#include <fstream>
#include <chrono>
#include <deque>
#include <condition_variable>

using namespace std;

std::deque<int> data_container;

std::mutex mu;

std::condition_variable cond;



// Using conditional variable and mutex
/*
    adding 10, 9 , 8, ... 1 to data_container.
*/
void func_generate_data() {
    
    int count = 10;

    while (count > 0) {
        
        //======================================
        // we use unique_lock since we just want to lock partial piece of code
        // ie. from locker(mu) to locker.unlock()
        // once a thead comess in, locker prevents other thread going through,
        // and they have to wait.
        std::unique_lock<mutex> locker(mu); 
        data_container.push_front(count);
        locker.unlock();
        //========================================


        //=========================================
        // use condition_variable to send out notification.
        cond.notify_one();  // Notify one waiting thread, if there is one.
        // cond.notify_all()
        //=========================================

        // wait 1 seconds.
        std::this_thread::sleep_for( chrono::seconds(1) );
        
        count--;
    }
}

/*
    using a while loop to continuously check data_container,
    once the read data = 1, break the loop.
*/
void func_read_data() {
    
    int data = 0;

    // note: data = 1 is the last element in data_container.
    while ( data != 1) {
        
        // =================================================
        std::unique_lock<mutex> locker(mu);
        

        cond.wait( locker, [](){ return ! data_container.empty();} );  
        // the first argument is locker,
        // which is to unlock the "mu" and then "sleep"
        // until a notification comes, cond will wake up
        // then go to next statement.
        // but cond can wait up ( not sleep forever )
        // this is called spurious wakeup.
        // when it wakes up, we can check 
        // so it comes the second argument,
        // it returns a true/false boolean
        // once data_container is empty (return false)
        // the cond continue to sleep ( not to wake it up)

         
        data = data_container.back();
        data_container.pop_back();    // remove the element at the back.

        locker.unlock();
        // =================================================

        cout << "t2 got a value from t1: " << data << endl;
    }
}




int main() {

    std::thread t1( func_generate_data );
    std::thread t2( func_read_data );
    t1.join();
    t2.join();
    return 0;
}
