/*
    part1:
    demo how to use std::async function (which creates a thread) and 
    std::future<T> to receive return value with type T from a thread.

    these two simplified the situations when we need the return values 
    from a thread.

    when using std::async(  std::launch::async, factorial, 6 );
    we will create a new thread by "std::launch::async"


    part2:
    demo how to use std::promise<T> to pass value to a thread.
    again, std::future<T> is our channel, for both return and argument.
    for argument, we need to link a future obj with a promise obj, by
    std::promise<T>::get_future(), 
    we then use std::promise<T>::set_value().

    note: using std::async() function, you are forced to use 
    a std::future<T> object to receive its return.

********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>     
#include <iostream>
#include <future>

using namespace std;

int factorial( int N ){

    int result = 1;
    for( int i=N; i>1; i--){
        result = result * i; 
    }

    printf("factorial %d! = %d\n", N, result ); 

    return result;
} 


int factorial_with_future( std::future<int>& fu ){
    // instead of passing an int for N,
    // we passing future object with <int>
    // and we use fu.get() to set the value for N.

    int N = fu.get();

    int result = 1;
    for( int i=N; i>1; i--){
        result = result * i; 
    }

    printf("factorial %d! = %d\n", N, result ); 

    return result;

}



int factorial_with_shared_future( std::shared_future<int> fu ){
    // instead of passing an int for N,
    // we passing shared_future object with <int>
    // note: it doesn't need to be reference, since 
    // shared_future obj can be copied.

    int N = fu.get();

    int result = 1;
    for( int i=N; i>1; i--){
        result = result * i; 
    }

    printf("factorial %d! = %d\n", N, result ); 

    return result;

}

 
int main(){

{
    // part 0 : intro

    // thread t1( factorial, 4 ); 
    // t1.join();
    /*  printing is fine, but what if we need to get the return from
        factorial function?
        (1) not a good way:
            create the variable x inside main(),
            then pass it to t1 by std::ref(x)
            here, we then require mutex to lock.

        (2) better way:
            use std::async function (which creates a thread)
            and use std::future<int> to receive return value.
    */ 
}




{
    // part 1: receive return value.

    std::future<int> fu = std::async( factorial, 4 );
    
    int x = fu.get(); // note: get() can only called once.

    printf("in main(), return value %d.\n", x);


    // note: std::async( factorial, 4 ); = 
    // std::async(  std::launch::async | std::launch::deferred, factorial, 4 );
    // depending on implementaiton, it may or may not create a new thread.


    // ======== let's see other case of application =============

    std::future<int> fu2 = std::async(  std::launch::deferred, factorial, 5 );
    
    // by adding std::launch::deferred, the factorial() will be executed when
    // fu2.get() is called., but no new threads are created, 
    // factorial() is executed in the main thread. 

    int x2 = fu2.get(); 

    printf("in main(), return value %d.\n", x2);

    std::future<int> fu3 = std::async(  std::launch::async, factorial, 6 );
    
    // by adding std::launch::async, there will be a new thread.

    int x3 = fu3.get(); 

    printf("in main(), return value %d.\n", x3);
}



{
    // part 2: passing value to a thread

    std::promise<int> prom; 
    std::future<int>  fu_arg  = prom.get_future(); 
    // link future<int> with promise<int>

    std::future<int> fu4 = std::async( std::launch::async, 
        factorial_with_future, std::ref( fu_arg ) );

    prom.set_value( 7 );
    int x4 = fu4.get(); 

    printf("in main(), return value %d.\n", x4);


    // if you cannot really do set_value(), ie. broken promise.
    // you should raise exception:
    // prom.set_exception( std::make_exception_ptr(
    //   std::runtime_error("To error is human")
    // ))

    // also note: one cannot pass the same std::future<T> to different thread
    // since std::future<T>::get() can only be executed once.
    // here comes the solution of shared_future.

    std::promise<int> prom2; 
    std::future<int>  fu_arg2  = prom2.get_future(); 
    std::shared_future<int> sf = fu_arg2.share(); // create a shared_future<T> obj.

    std::future<int> fu5 = std::async( std::launch::async, factorial_with_shared_future, sf );
    std::future<int> fu6 = std::async( std::launch::async, factorial_with_shared_future, sf );
    std::future<int> fu7 = std::async( std::launch::async, factorial_with_shared_future, sf );

    prom2.set_value( 2 );
}


    return 0;
}
