
#include "B1DetectorConstruction.hh"
#include "B1ActionInitialization.hh"

#ifdef G4MULTITHREADED
#include "G4MTRunManager.hh"
#else
#include "G4RunManager.hh"
#endif


#include "Randomize.hh"
#include "QBBC.hh"           // physics list

#include "G4VisExecutive.hh"

#include "G4UImanager.hh"     // User Interface manager
#include "G4UIExecutive.hh"   // for command terminal or macro input.




/*!
    @brief Main program of the B1 example

    we create a runManager either from G4RunManager or G4MTRunManager.
    runManager will use objects for detector construction, phyiscs process, 
    and the user actions.
    
    we create a visManager.
    we create a UImanager.

    use UImanager to start session.
    when we finish, delete all the managers.
*/
int main(int argc,char** argv)
{


  // Detect interactive mode (if no arguments) and define UI session
  G4UIExecutive* ui = 0;
  if ( argc == 1 ) { ui = new G4UIExecutive(argc, argv); }




  // Choose the Random engine
  G4Random::setTheEngine(new CLHEP::RanecuEngine);



  // Construct the default run manager
#ifdef G4MULTITHREADED
  G4MTRunManager* runManager = new G4MTRunManager; 
#else
  G4RunManager* runManager = new G4RunManager;
#endif




  // Detector construction
  runManager->SetUserInitialization(new B1DetectorConstruction());


  // Physics list
  G4VModularPhysicsList* physicsList = new QBBC;
  physicsList->SetVerboseLevel(1);
  // VerboseLevel is  related to print out. We have 0 to 4 levels.
  runManager->SetUserInitialization(physicsList);


  // User action initialization
  runManager->SetUserInitialization(new B1ActionInitialization());







  // Initialize visualization
  //
  G4VisManager* visManager = new G4VisExecutive;
  // G4VisExecutive can take a verbosity argument - see /vis/verbose guidance.
  // G4VisManager* visManager = new G4VisExecutive("Quiet");
  visManager->Initialize();
  // and we need to delete visManager in the end.




  // Get the pointer to the User Interface manager
  G4UImanager* UImanager = G4UImanager::GetUIpointer();

  // Process macro or start UI session
  if ( ! ui ) {
    // batch mode
    G4String command = "/control/execute ";
    G4String fileName = argv[1];
    UImanager->ApplyCommand(command+fileName); // takeing macro from command line.
  }
  else {

    // interactive mode
    UImanager->ApplyCommand("/control/execute init_vis.mac");

    ui->SessionStart();
    //
    //  then, we can type in command manually.
    //
    delete ui;
  }

  // Job termination
  // Free the store: user actions, physics_list and detector_description are
  // owned and deleted by the run manager, so they should not be deleted
  // in the main() program !

  delete visManager;
  delete runManager;
}


