//file B1PrimaryGeneratorAction.cc
// brief Implementation of the B1PrimaryGeneratorAction class

#include "B1PrimaryGeneratorAction.hh"

#include "G4LogicalVolumeStore.hh"
#include "G4LogicalVolume.hh"
#include "G4Box.hh"
#include "G4RunManager.hh"
#include "G4ParticleGun.hh"
#include "G4ParticleTable.hh"
#include "G4ParticleDefinition.hh"
#include "G4SystemOfUnits.hh"
#include "Randomize.hh"

/*! @brief ### set up our gamma beam
    It is 6MeV gamma, moving toward (0,0,1) direction. <br>
    gamma's location will be set in GeneratePrimaries() <br>

   + run the G4VUserPrimaryGeneratorAction() constructor.
   + initial the fParticleGun = 0, fEnvelopeBox =0
*/
B1PrimaryGeneratorAction::B1PrimaryGeneratorAction()
    : G4VUserPrimaryGeneratorAction(),
      fParticleGun(0),
      fEnvelopeBox(0)
{

    G4int n_particle = 1;

    fParticleGun  = new G4ParticleGun(n_particle);


    // set up a gamma particle.
    G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
    G4String particleName;
    G4ParticleDefinition* particle
        = particleTable->FindParticle(particleName="gamma");


    // put the gamma particle, momentum direction, energy
    // into the fParticleGun obj.
    fParticleGun->SetParticleDefinition(particle);
    fParticleGun->SetParticleMomentumDirection(G4ThreeVector(0.,0.,1.));
    fParticleGun->SetParticleEnergy(6.*MeV);
}






///  we must delete the fParticleGun when the B1PrimaryGeneratorAction obj is deleted.
B1PrimaryGeneratorAction::~B1PrimaryGeneratorAction()
{
    delete fParticleGun; 
}




/*!
    @brief ###setup the gamma's location <br>
    to get the width and height of the 'Envelope' phyisclogic volume. <br>
    we set 0.8*width and 0.8*height as the area that the gamma will hit. <br>
*/
void B1PrimaryGeneratorAction::GeneratePrimaries(G4Event* anEvent)
{
    /*! + this is called at the begining of ecah event 
        + In order to avoid dependence of DetectorConstruction class,
        we get Envelope volume from G4LogicalVolumeStore.
    */



    G4double envSizeXY = 0;
    G4double envSizeZ = 0;


    //  to retrive the 'Envelope' G4Box obj to our fEnvelopeBox obj.

    if (!fEnvelopeBox) // note: initially fEnvelopeBox = 0 = false, so !false = true
    {
        G4LogicalVolume* envLV
            = G4LogicalVolumeStore::GetInstance()->GetVolume("Envelope");

        if ( envLV ) fEnvelopeBox = dynamic_cast<G4Box*>(envLV->GetSolid());
    }



    if ( fEnvelopeBox )
    {
        // if the retrival success, then  assign the width and length to our envSizeXY and envSizeZ variables.

        envSizeXY = fEnvelopeBox->GetXHalfLength()*2.;

        envSizeZ = fEnvelopeBox->GetZHalfLength()*2.;
    }
    else
    {
        // However, if not success, then send out an error message. ( and an exception.)

        G4ExceptionDescription msg;
        msg << "Envelope volume of box shape not found.\n";
        msg << "Perhaps you have changed geometry.\n";
        msg << "The gun will be place at the center.";
        G4Exception("B1PrimaryGeneratorAction::GeneratePrimaries()",
                    "MyCode0002",JustWarning,msg);
    }


    /*
    | set the active area = 0.8 width * 0.8 length
    | and use G4UniformRand to control the gamma ray position.
    */
    G4double size = 0.8;
    G4double x0 = size * envSizeXY * (G4UniformRand()-0.5);
    G4double y0 = size * envSizeXY * (G4UniformRand()-0.5);
    G4double z0 = -0.5 * envSizeZ;

    // set up the gamma ray position by G4ThreeVector obj.
    fParticleGun->SetParticlePosition(G4ThreeVector(x0,y0,z0));




    // anEvent is our input argument.
    // data type note: G4ParticleGun*  fParticleGun;
    // think of this line is a routine.
    fParticleGun->GeneratePrimaryVertex(anEvent);



}


