/// @file B1RunAction.cc
/// @brief Implementation of the B1RunAction class

#include "B1RunAction.hh"
#include "B1PrimaryGeneratorAction.hh"
#include "B1DetectorConstruction.hh"
// #include "B1Run.hh"

#include "G4RunManager.hh"
#include "G4Run.hh"
#include "G4ParameterManager.hh"
#include "G4LogicalVolumeStore.hh"
#include "G4LogicalVolume.hh"
#include "G4UnitsTable.hh"
#include "G4SystemOfUnits.hh"

/*!
  + define units (milligray, microgray, ... )
  + Register our fEdep and fEdep2 parameters by parameterManager.
*/
B1RunAction::B1RunAction()   : 
    G4UserRunAction(), 
    fEdep("Edep", 0.),  
    fEdep2("Edep2", 0.)
    
{

  // add new units for dose
  const G4double milligray = 1.e-3*gray;
  const G4double microgray = 1.e-6*gray;
  const G4double nanogray  = 1.e-9*gray;
  const G4double picogray  = 1.e-12*gray;


  //		      (  name     ,  symbol   , category, value )
  new G4UnitDefinition("milligray", "milliGy" , "Dose", milligray);
  new G4UnitDefinition("microgray", "microGy" , "Dose", microgray);
  new G4UnitDefinition("nanogray" , "nanoGy"  , "Dose", nanogray);
  new G4UnitDefinition("picogray" , "picoGy"  , "Dose", picogray);

  // Register parameter to the parameter manager
  G4ParameterManager* parameterManager = G4ParameterManager::Instance();
  parameterManager->RegisterParameter(fEdep);
  parameterManager->RegisterParameter(fEdep2);
}

/// do nothing..
B1RunAction::~B1RunAction() {}




/*!
  + BeginOfRunAction() will be invoked at the begining of BeamOn()
  + set to not record random number.
  + call parameterManager to reset all registered parameters to zero.
*/
void B1RunAction::BeginOfRunAction(const G4Run*)
{
   

  // inform the runManager to save random number seed
  G4RunManager::GetRunManager()->SetRandomNumberStore(false);

  // reset parameters to their initial values
  G4ParameterManager* parameterManager = G4ParameterManager::Instance();

  parameterManager->Reset(); // set fEdep and fEdep2 to 0 

}






/*!
  + do parameter merging ( for multi thread )
  + from fEdep and fEdep2, calculate rms.
  + from rms, calculate rms/dos
  + print out data on the screen ( only for sequence mode )

*/
void B1RunAction::EndOfRunAction(const G4Run* run)
{
  G4int nofEvents = run->GetNumberOfEvent();
  if (nofEvents == 0) return;

  // Merge parameters for multi-thread purpose.
  G4ParameterManager* parameterManager = G4ParameterManager::Instance();
  parameterManager->Merge();




  // Compute dose = total energy deposit in a run and its variance
  G4double edep  = fEdep.GetValue();
  G4double edep2 = fEdep2.GetValue();

  G4double rms = edep2 - edep*edep/nofEvents;
  if (rms > 0.) rms = std::sqrt(rms); else rms = 0.;



  // retrieve our detectorConstruction
  const B1DetectorConstruction* detectorConstruction
   = static_cast<const B1DetectorConstruction*>
     (G4RunManager::GetRunManager()->GetUserDetectorConstruction());



  G4double mass = detectorConstruction->GetScoringVolume()->GetMass();
  G4double dose = edep/mass;
  G4double rmsDose = rms/mass;





  // Run conditions
  //  note: There is no primary generator action object for "master"
  //        run manager for multi-threaded mode.

  const B1PrimaryGeneratorAction* generatorAction
   = static_cast<const B1PrimaryGeneratorAction*>
     (G4RunManager::GetRunManager()->GetUserPrimaryGeneratorAction());



  G4String runCondition;

  if (generatorAction)
  {
    // if the retrival is ok.

    const G4ParticleGun* particleGun = generatorAction->GetParticleGun();

    runCondition += particleGun->GetParticleDefinition()->GetParticleName();

    runCondition += " of ";

    G4double particleEnergy = particleGun->GetParticleEnergy();

    runCondition += G4BestUnit(particleEnergy,"Energy");
  }








  // Print
  //
  if (IsMaster()) {
    G4cout
     << G4endl
     << "--------------------End of Global Run-----------------------";
  }
  else {
    G4cout
     << G4endl
     << "--------------------End of Local Run------------------------";
  }

  G4cout
     << G4endl
     << " The run consists of " << nofEvents << " "<< runCondition
     << G4endl
     << " Cumulated dose per run, in scoring volume : "
     << G4BestUnit(dose,"Dose") << " rms = " << G4BestUnit(rmsDose,"Dose")
     << G4endl
     << "------------------------------------------------------------"
     << G4endl
     << G4endl;
}



/*!
  @brief for a user event action object to update fEdep and fEdep2. <br>
  fEdep represents the energy deposit of a run, which consists many events.

  @param edep
  the deposit energy from a given event.  
*/
void B1RunAction::AddEdep(G4double edep)
{
  // note: G4Parameter<G4double> fEdep;
  fEdep  += edep;
  fEdep2 += edep*edep;

}



