import ROOT
import scipy
from scipy import integrate
import numpy as np
from numpy.random import uniform
import math



def pdfFunction( x, *args):
    ''' the probability distribution function used in get_intg() 
        we perform linear interpolation via TGraph::Eval().

        param
        ------
        x => float 
        *agrs => argument list. only the first one will be used.
        It takes in the TGraph object for angular distribution.
         
    '''
    gr_pdf = args[0]
    
    return gr_pdf.Eval( x )

def get_intg( pdfFunction, xmin, xmax, gr_pdf ):
    ''' perform an integration from [xmin, xmax] according to pdfFunction()
        using Gaussian quadrature method, whose speed is much faster 
        than brutal integration.

        param
        -------
        pdfFunction   => function.
        xmin => float, the lower limit of an integration.
        ang    => float, the upper limit of an integration.
        gr_pdf => TGraph object.

        return
        ------
        integration value. 
    '''
    result,err = scipy.integrate.quad( pdfFunction, xmin, xmax, args=(gr_pdf,)  )

    return result

#=========================================================
# cdf

def get_CDF( gr_pdf, xmin, xmax, pointN = 100 ):
    ''' prepare the TGraph object for cdf -- cumulative distribution function

        param
        -----
        gr_pdf, TGraph object
        xmin, float
        xmax, float

        note
        -----
        to do cdf, we need normalization, and thus, we need xmin and xmax.
        
        due to numerical issue, we need to take correcting the
        integration artificially.

        return
        ------
        a TGraph object for cdf. 
        
    '''

   
    norm = get_intg( pdfFunction, xmin, xmax, gr_pdf )

    
    gr = ROOT.TGraph(); ipt = 0
    
    preValue = 0
    isIncre = None  # cdf should be alwasy increasing.
    
    
    for xx in np.linspace( xmin, xmax, pointN ):
    

        intg = get_intg( pdfFunction, xmin, xx , gr_pdf )
        intg = intg/norm 
         
        #=============================================
        # make cdf alwasy increasing and <= 1
        if( intg-preValue <= 0  and (preValue!= 0) ):
            intg = preValue + (1. - preValue ) * 0.001
    
        # cdf's max is 1. 
        if( intg > 1. ): 
            intg = preValue + (1. - preValue ) * 0.001
        
        preValue = intg 
        #============================================== 
        
        gr.SetPoint( ipt, xx, intg )            
        
        ipt += 1
 
    preValue = 0
    showWarnning = False
    for ipt in range( gr.GetN() ):
        x = ROOT.Double( 0. )
        y = ROOT.Double( 0. )
        gr.GetPoint( ipt, x, y )
        if y < preValue: showWarnning = True 
        preValue = y 

    if( showWarnning ):
        print "WARNING: cdf is not an increasing function."
    return gr    



#=========================================================
# inverse cdf

def get_inverseCDF( gr_cdf ):
    ''' from gr_cdf prepare its inverse function, gr_inverseCDF,

        param
        ------
        gr_cdf => TGraph object, prepared by get_CDF()

        return
        ------
        gr_inverseCDF,
        ymin, float, should be close to 0, related to xmin. (just for diagnose)
        ymax, float, should be close to 1, related to xmax. (just for diagnose)
        
    '''
    x = ROOT.Double( 0.)
    y = ROOT.Double( 0.)

    gr = ROOT.TGraph()
    for ipt in range( gr_cdf.GetN() ):
        gr_cdf.GetPoint( ipt, x, y )
        gr.SetPoint( ipt, y, x )
       

    ymin = ROOT.TMath.MinElement( gr.GetN(),gr.GetX() ); 
    ymax = ROOT.TMath.MaxElement( gr.GetN(),gr.GetX() ); 
    return gr, ymin, ymax;   


def gr_normalized( grData, norm ):
    ''' normalizing a TGraph object by norm 
        param
        -------
        grData => TGraph type
        norm   => float type

        return => TGraph object.
    '''
    x = ROOT.Double( 0. )
    y = ROOT.Double( 0. )

    gr = ROOT.TGraph()
    for ipt in range( grData.GetN() ):

        grData.GetPoint( ipt, x, y )
        gr.SetPoint( ipt, x, y/norm )

    return gr


def get_norm( gr, x1, x2 ):
    '''
    param
    ======
    gr, TGraph, the object to be normalized.
    x1 and x2 is the range that we will normalize gr to.
    '''
    norm = 0.
    stepN = 20

    for xx in np.linspace( x1, x2, stepN  ):
        norm += gr.Eval( xx )

    return norm/stepN 


def get_sampling_INV( gr_inverseCDF, xmin, xmax, ymin, ymax  ):
    ''' given a inverseCDF function, 
        we random sample 10^6 events for yy value between ymin and ymax, 
        should be between (0,1).

        use xx = inverseCDF( yy ), fill xx into histo.
        convert histo to gr object.

        return
        ------
        TGraph object, (x-axis) ang (y-axis) the sampled angular distribution, 
        not normalized yet.
    '''
    
    histo = ROOT.TH1D( "histo", "histo", 1000, xmin, xmax )

    for ix in range( 1000000  ):
        yy = ymin + np.random.uniform(0,1) * (ymax-ymin)
        xx = gr_inverseCDF.Eval( yy )
        histo.Fill( xx )


    gr = ROOT.TGraph()
    ipt = 0
    for xx in np.linspace( xmin, xmax, 1000 ):
        xbin = histo.GetXaxis().FindBin( xx )
        value = histo.GetBinContent( xbin )
        gr.SetPoint( ipt, xx, value ) 
        ipt +=1
    
    return gr




 
def get_sampling_REJ( gr_pdf, xmin, xmax ):
    ''' using rejection method to sample gr_pdf 
        sampleN = 10^6.
    '''

    histo = ROOT.TH1D( "histo", "histo", 1000, xmin, xmax )

    # get ymax and ymin
    ymax = -99999999
    ymin =  99999999
    for xx in np.linspace( xmin, xmax, 1000 ):
        yy = gr_pdf.Eval( xx )
        if( yy > ymax ): ymax = yy
        if( yy < ymin ): ymin = yy

    iCount = 0
    while( 1 ):
        xr = xmin + np.random.uniform(0,1) * (xmax-xmin)
        yr =        np.random.uniform(0,1) * ymax
        # note: the yr is sample from 0 to ymax

        if( gr_pdf.Eval(xr) > yr ):
            histo.Fill( xr )
            iCount += 1
        if( iCount > 1000000 ): break
    

    gr = ROOT.TGraph()
    ipt = 0
    for xx in np.linspace( xmin, xmax, 1000 ):
        xbin = histo.GetXaxis().FindBin( xx )
        value = histo.GetBinContent( xbin )
        gr.SetPoint( ipt, xx, value ) 
        ipt +=1
    
    return gr


 
    

grs = []
## give an gr_pdf, and the range for xmin and xmax,
## we return a gr_cdf 
if __name__ == '__main__':
    
    #==========================================================
    # user input
    # pdf 
    gr_pdf = ROOT.TGraph()
    gr_pdf.SetPoint( 0,  0, 8.8e-04 );
    gr_pdf.SetPoint( 1, 10, 8.9e-04 );
    gr_pdf.SetPoint( 2, 20, 8.9e-04 );
    gr_pdf.SetPoint( 3, 30, 8.9e-04 );
    gr_pdf.SetPoint( 4, 40, 8.7e-04 );
    gr_pdf.SetPoint( 5, 50, 8.3e-04 );
    gr_pdf.SetPoint( 6, 60, 7.3e-04 );
    gr_pdf.SetPoint( 7, 70, 6.3e-04 );

    gr_pdf.SetPoint( 8, 90, 4.4e-04 );
    gr_pdf.SetPoint( 9, 110, 3.1e-04 );
    gr_pdf.SetPoint( 10, 130, 2.3e-04 );
    gr_pdf.SetPoint( 11, 150, 1.9e-04 );
    gr_pdf.SetPoint( 12, 180, 1.7e-04 );

    xmin = 0.
    xmax = 180.
    #==========================================================
    
    gr_cdf = get_CDF( gr_pdf, xmin, xmax, pointN=200 ) # cdf
    
    gr_inverseCDF, ymin, ymax = get_inverseCDF( gr_cdf )




    #==========================================================
    ## validation
    norm   = gr_pdf.Eval( xmin )
    gr_pdf_norm = gr_normalized( gr_pdf, norm )    
    grs.append( gr_pdf_norm )
        

    gr_sampleINV = get_sampling_INV( gr_inverseCDF, xmin, xmax, ymin, ymax )
    norm   = get_norm( gr_sampleINV, xmin, xmin+10 )
    gr_sampleINV = gr_normalized( gr_sampleINV, norm ) 
    grs.append( gr_sampleINV )


  
    # gr_sampleREJ = get_sampling_REJ( gr_pdf, xmin, xmax )
    # norm   = get_norm( gr_sampleREJ, xmin, xmin+10 )
    # gr_sampleREJ = gr_normalized( gr_sampleREJ, norm ) 
    # grs.append( gr_sampleREJ )

    # plot
    canvas1 = ROOT.TCanvas("canvas1")
    canvas1.SetGridx(1); canvas1.SetGridy(1);
    # canvas1.SetLogy(1);

    mg = ROOT.TMultiGraph()
    for i, gr in enumerate( grs ) :  
        gr.SetLineWidth(2)
        gr.SetLineColor( i+1 )
        gr.SetLineStyle( i+1 )
        mg.Add( gr )


        mg.Draw("AL")
        canvas1.Update()
        mg.GetYaxis().SetRangeUser( 0, 1.1 )
 
        