#include "Histo1D.hpp"
#include <iostream>
#include <fstream>
#include <cmath>
#include <limits>

// constructor
Histo1D::Histo1D( int binN, double xmin, double xmax ) :
    fXmin( xmin ), fXmax( xmax ), fBinN( binN )
{

     int totalBinN = binN + 2; 
    // data from idx 1 to binN, 
    // idx = 0, underflow, idx=binN+1 = overflow.

    fData = new double [ totalBinN ];

    for( int i = 0; i < totalBinN; i++ ){ fData[i] = 0.; }

    fStepSize = (  xmax - xmin )/ binN; 
}



// destructor
Histo1D::~Histo1D(){

    delete [] fData;
}



void Histo1D::Fill( double value ){

    int xbin = FindBin( value );

    fData[ xbin ] = fData[ xbin ] + 1;
}

double Histo1D::GetBinCenter( int xbin ){

    double center = xbin*fStepSize - 0.5*fStepSize + fXmin;

    return center;
}


double Histo1D::GetBinContent( int xbin ){

    return fData[ xbin ];
}



double Histo1D::GetContentAtXX( double xx ){

    return  GetBinContent(  FindBin(xx) );

}



void   Histo1D::SetBinContent( int xbin, double value ){

    fData[ xbin ] = value;
}




int Histo1D::FindBin( double value ){


    int idx =  int( (value-fXmin)/fStepSize + 0.00001 ) + 1;

    if( value < fXmin )      { return 0; }
    else if( value > fXmax ) { return fBinN + 1 ; }
    else { return idx; }
    
}




std::vector<int>& Histo1D::FindBin_maxFreq() {

    double ymax = -1;

    std::vector<int>* v_bin  = new std::vector<int>();

    for( int iBin = 1; iBin<=fBinN; iBin++ ){
         
         if( fData[iBin] > ymax){
            
            ymax = fData[iBin];

            v_bin->clear();
            v_bin->push_back( iBin );

         } else if( fData[iBin] == ymax ) {

            v_bin->push_back( iBin );
         }
    }

    return *v_bin;
}



void Histo1D::Print( ){

    for( int iBin = 1; iBin<=fBinN; iBin++ ){
        printf("bin %5d, value = %15lf, center = %15lf\n", 
            iBin, fData[iBin], GetBinCenter(iBin) );
    }
}




void Histo1D::Clear( ){
     for( int i = 0; i < fBinN+2; i++ ){ fData[i] = 0.; }
}



void Histo1D::Compare(  Histo1D & histo2 )  
{

    printf("%10s, %10s, %10s %10s %10s %10s\n", 
        "bin", "h1", "h2", "diff", "dratio", "center" );

    for( int iBin = 1; iBin<=fBinN; iBin++ ){

        double h2Value = histo2.GetBinContent(iBin);
        double diff = fabs( fData[iBin] - h2Value );
        double rdiff = diff/fData[iBin]; 
       
        // test diff
        // printf("h1 =%10.2f, h2=%10.2f, diff = %.10e\n", 
        //     fData[iBin], h2Value, diff);

        printf("%10d, %10.2lf, %10.2lf, %10.2lf, %10.2lf, %10lf\n", 
        iBin, fData[iBin], h2Value,
        diff, rdiff,
        GetBinCenter(iBin) );
    }

    printf("h1sum = %Lf h2sum = %Lf\n", GetSum(), histo2.GetSum() );
}




void Histo1D::Normalize( double norm ){

    for( int i = 0; i < fBinN+2; i++ ){ fData[i] *= norm; }
}



long double Histo1D::GetSum(){

    long double sum = 0;

    for( long int i = 0; i < fBinN+2; i++ ){ 
        sum += fData[ i ];
    }

    return sum;
}



void Histo1D::WriteOut( const char* fName ){

    std::fstream fout;

    fout.open( fName, std::ios::out );

    for( int ii = 1; ii <= 1000; ii ++ )
    {
        fout << fData[ii] << "\n"; 
    }
}


double Histo1D::GetStandardDeviation(){

    // find the avg center
    double avg = 0.;
    double total_w = 0.;
    for (int ibin = 1; ibin <= fBinN; ++ibin )
    {
        double w = GetBinContent( ibin );
        total_w += w ; 
        avg += w * GetBinCenter( ibin );    
    
    }
    avg = avg / total_w;
   

    // find the variance
    double var = 0.;
    for (int ibin = 1; ibin <= fBinN; ++ibin)
    {
        double w = GetBinContent( ibin );
        double value = GetBinCenter( ibin );    
        var += w * (avg-value )*( avg-value ) ;
    }
    var = var / (total_w-1);
    
    return sqrt( var );
}



double Histo1D::PrintRandomness(){

    // find the avg, max, min
    double v_max = std::numeric_limits<double>::lowest();
    double v_min = std::numeric_limits<double>::max();
    double avg = 0.;
    double total_w = 0.;
    for (int ibin = 1; ibin <= fBinN; ++ibin )
    {
        double w = GetBinContent( ibin );
        double value = GetBinCenter( ibin );
        total_w += w ; 
        avg += w * value;    
        if( w > v_max ){ v_max = w; }
        if( w < v_min ){ v_min = w; }
    }
    avg = avg / total_w;

    

    // find the variance
    double var = 0.;
    for (int ibin = 1; ibin <= fBinN; ++ibin)
    {
        double w = GetBinContent( ibin );
        double value = GetBinCenter( ibin );    
        var += w * (avg-value )*( avg-value ) ;
    }
    var = var / (total_w-1);

    // standard deviation

    double stdev = sqrt( var ) / avg;

    printf("stdev/avg = %10.3e, (max-min) = %10.3e, max = %10.3e, min = %10.3e \n\n", 
        stdev, (v_max-v_min), v_max, v_min );
    
    return 1./(v_max-v_min)/stdev;

}