/// @file Histo1D.hpp

#ifndef _RB_HISTO1D_
#define _RB_HISTO1D_

#include <vector>
 
/*! @brief a light-weight class 1D histogram, currently only 
    support the xmin and xmax to be double type, and the values
    in the bins are also double.

    The design resembles the TH1 class in ROOT, 
    e.g. GetBinCenter(), GetBinContent(), etc..
    while there are some new member functions, like
    GetContentAtXX(), FindBin_maxFreq()....
    This class is used to check the sampling for ang. distribution,
    and the random generators.
*/ 
class Histo1D {

private:
 
    double* fData;  ///< data array.
    double  fXmin;  ///< histo's xmin
    double  fXmax;  ///< histo's xmax
    double  fBinN;  ///< histo's bin number
    double  fStepSize; ///< histo's bin size = (xmax-xmin)/binN

public:
    
    /*! @brief constructor
        
        all contents in the bins are initialized to 0.

        @param binN = number of bins
        @param xmin = min x range.
        @param xmax = max x range.
    */
    Histo1D( int binN, double xmin, double xmax ); 
    
    Histo1D(){;} // do nothing.

    /*! @brief destructor
        delete the fData
    */ 
    ~Histo1D();


	/*! @brief Fill one count to the bin which corresponding to the input value.
		
		@note when value < xmin, then one count is filled in the bin0,
		which is for the underflow bin, when value >= xmax, then one count 
		is filled in the binN+1 position, which is the overflow bin.
	*/
    void Fill( double value );


	/*! @brief return the center position for a given bin 
    */
    double GetBinCenter( int xbin );


	/*! @brief return the content (i.e. frequency) for a given bin 
    */
    double GetBinContent( int xbin );


	/*! @brief return the center content (i.e. frequency) for a position,
		i.e. position will be turned into bin number first, then 
		use the bin number to get the content of a given bin.
	*/
    double GetContentAtXX( double xx );

	
	/*! @brief set content (value) to a given bin (xbin) 
    */
    void   SetBinContent( int xbin, double value );


	/*! @brief return xbin position which relates a given value 
    */
    int    FindBin( double value );
  
	
	/*! @brief return xbin positions which corresponds the max frequency.  

		@return a std::vector obj. 

		@note there could be multiple positions have the same max value.
	*/
    std::vector<int>&   FindBin_maxFreq();


	/*! @brief print basic info for the histo 
    */
    void   Print( );


	/*! @brief reset the contents for the all bins to 0.  
    */
    void   Clear( );


	
	/*! @brief for each bin, multiple norm to its content.  
    */
    void   Normalize( double norm );


	/*! @brief print out the comparison between this histo and the input histo. 
    */
    void   Compare( Histo1D & obj ) ;


	/*! @brief write out each bin content to an external file. 
		@param fName is the name of the file to be written out.
	*/
    void   WriteOut( const char* fName );


	/*! @brief get the sum of the contents from all bins.  
    */
    long double GetSum();


    /*! @brief get standard deviation from all the range (i.e. bin1 to binN)
    */
    double GetStandardDeviation();

    /*! @brief print the stdev, avg, and (v_max-vmin)
    */
    double PrintRandomness();
    
};
#endif

  
